#include "stdafx.h"

//=========================================================================
// File: klx_test.cpp
//		klx̊eCũeXgR[h
//=========================================================================

#include "metaprog.hpp"
#include "smartc.hpp"
#include "opfun.hpp"
#include "binary_c.hpp"
#include "lengthof.hpp"
#include "initcont.hpp"
#include "scopeguard.hpp"



namespace klx
{
	namespace
	{
		struct counter
		{
			static void up(int=0)   { ++count; }
			static void down(int=0) { --count; }
			static void reset()     { count=0; }
			static int count;
		};
		int counter::count = 0;
	}

	static void metaprog_unittest()
	{
		assert( 100 == sizeof(sized_type<100>::type) );
		assert( 10 != sizeof(sized_type<9>::type) );
		assert( sizeof(yes_type) != sizeof(no_type) );
		int hoge(other_type);
		assert( sizeof(hoge(1), hoge("hello,world"), hoge(op_nothing())) );
	}

	static void smartc_unittest()
	{
		{
			smart_constant<int> x( 100, op_nothing() ), y( 200, op_nothing() );
			assert( x != y );
			assert( x == 100 );
			assert( x != 200 );
			assert( x < 200 );
			assert( x > 99 );
			assert( *y-*x == x );
			assert( *y-*x != y );
			assert( *y-*x <= x );
			assert( *y+*x >= y );
			x = y;
			assert( x == y );
		}
		{
			counter::reset();
			{
				smart_constant<int> x( 2, &counter::up ), y, z;
				assert( counter::count == 0 );
				y = x; z = x; z = y;
				assert( counter::count == 0 );
				x = x;
				assert( counter::count == 0 );
				assert( *y == 2 );
			}
			assert( counter::count == 1 );
			{
				smart_constant<int> x( 2, &counter::up ), y;
				assert( counter::count == 1 );
				x = y;
				assert( counter::count == 2 );
			}
		}
	}

	static void opfun_unittest()
	{
		assert( op_equal_to()(16, op_lshift()(4, op_plus()(1, op_modulus()(100,3)))) );
	}

	static void binaryc_unittest()
	{
		assert( _0b(1101) == 13 );
		assert( _0b2(00001111,11110000) == 0x0ff0 );
		int hoge[_0b(0101)];
		assert( sizeof(hoge) == sizeof(int)*5 );

		hoge[0]=1; // warning
	}

	static void lengthof_unittest()
	{
		int hoge[100], *p[200];
		assert( lengthof(hoge) == 100 );
		assert( lengthof(p) == 200 );
		double xxx[lengthof(hoge)];
		assert( lengthof(xxx) == 100 );

		p[1] = hoge; // warning
		xxx[0] = 0; // wanning
	}

	static void initcont_unittest()
	{
		vector<int> x;
		init_v(x) = 1,2,3,4,5,6,7,8,9;
		addto_v(x) = 10,11,12,13;
		assert( x.size() == 13 );
		for(unsigned i=0; i!=x.size(); ++i)
			assert( x[i] == int(i+1) );

		set<int> y;
		addto_s(y) = 15,25;
		init_s(y) = 10,20,30,40;
		assert( y.size() == 4 );
		assert( y.count(20) == 1 );
		assert( y.count(15) == 0 );

		stack<int> z;
		addto_q(z) = 1,2,3,4,5;
		assert( z.top() == 5 );
		z.pop();
		assert( z.top() == 4 );
	}

	static void scopeguard_unittest()
	{
		counter::reset();
		{
			counter::up();
			ON_BLOCK_END( &counter::down, 0 );
		}
		assert( counter::count == 0 );
		{
			counter::up();
			ON_EXCEPTION( &counter::down, 0 );
		}
		assert( counter::count == 1 );
		try
		{
			counter::up();
			ON_EXCEPTION( &counter::down, 0 );
			throw 1;
		} catch(int){}
		assert( counter::count == 1 );
	}

	void unittest()
	{
		metaprog_unittest();
		smartc_unittest();
		opfun_unittest();
		binaryc_unittest();
		lengthof_unittest();
		initcont_unittest();
		scopeguard_unittest();
	}
}
