#pragma once

//=========================================================================
// Package: klx::smart_constant
//=========================================================================

namespace klx {



//-------------------------------------------------------------------------
// Class: smart_constant<T>
//    '萔'ev[g
//
//	̃IuWFNgɋL萔lIuWFNgǗ܂B
//	v́AQƃJEgOɂȂ_œo^ꂽj֐ĂяoA
//	Ƃ́B̓IȗprƂĂ
//
//		- OS̕ԂnhbvĊǗ
//		- X}[g|C^ev[g̉x
//		- ȈՃXR[vK[h
//
//	Ȃǂl܂B
//
//		- zQƂ̓_
//		- Öق̕ϊ̓T|[gȂ
//		- OS
//		- XbhZ[t (#if build with _MT)
//
//	boost::shared_ptr(<http://www.boost.org/libs/smart_ptr/shared_ptr.htm>)
//  QlɁA|C^vfi[邽߂Ƀev[gƂč쐬܂B
//-------------------------------------------------------------------------

template<typename T>
class smart_constant
	: boost::less_than_comparable1< smart_constant<T>
	, boost::less_than_comparable2< smart_constant<T>, T
	, boost::equality_comparable1 < smart_constant<T>
	, boost::equality_comparable2 < smart_constant<T>, T
	> > > >
{
public:
	typedef smart_constant<T> this_type;
	typedef                T  element_type;

public:
	//
	// Function: smart_constant
	//	l̓ftHgBj֐
	//
	smart_constant()
		: body( new counted_body(T()) )
	{
	}

	//
	// Function: smart_constant
	//	lƔj֐𖾎w肵ď
	//
	template<typename D>
	explicit smart_constant( const T& t, D d )
		: body( new counted_body_with_deleter<D>(t,d) )
	{
	}

	//
	// Function: smart_constant
	//	Rs[
	//
	smart_constant( const this_type& rhs )
		: body( rhs.body->countUp() )
	{
	}

	//
	// Function: operator=
	//	
	//
	this_type& operator=( const this_type& rhs )
	{
		body->countDown();
		body = rhs.body->countUp(); // countUp  no throw 䂦̏ňS
		return *this;
	}

	~smart_constant()
	{
		body->countDown();
	}

public:
	//
	// Function: operator*
	//	i[Ăl̎擾
	//
	const T operator*() const
	{
		return body->getData();
	}

private:
	// QƃJEg@\̎
	#ifdef BOOST_HAS_THREADS
		#define KLX_SMARTC_COUNTUP(v) ::InterlockedIncrement(&v)
		#define KLX_SMARTC_COUNTDN(v) ::InterlockedDecrement(&v)
	#else
		#define KLX_SMARTC_COUNTUP(v) (++v)
		#define KLX_SMARTC_COUNTDN(v) (--v)
	#endif

	class counted_body
	{
	public:
		counted_body( const T& t ) : data(t), count(1) {}
		virtual ~counted_body() {}
		const T getData() const { return data; }
		counted_body* countUp() { KLX_SMARTC_COUNTUP(count); return this; }
		void countDown() { if( KLX_SMARTC_COUNTDN(count)==0 )delete this; }
	private:
		const T data;
		long   count;
	};

	#undef KLX_SMARTC_COUNTUP
	#undef KLX_SMARTC_COUNTDN

	// j֐Ăяo@\̎
	template<typename D>
	class counted_body_with_deleter : public counted_body
	{
	public:
		counted_body_with_deleter( const T& t, D d )
			: counted_body(t), deleter(d) {}
	private:
		virtual ~counted_body_with_deleter() { deleter(getData()); }
		D deleter;
	};

private:
	counted_body* body;
};



//-------------------------------------------------------------------------
// ⏕ZqB
//-------------------------------------------------------------------------

//
// Function: operator==
//		lr
//
//	smart_constant<T>  T ̔r`Ă܂B!=Zq́A
//	boost::operators ɂ莩܂B
//

template<typename T>
inline bool operator==
	( const smart_constant<T>& lhs, const smart_constant<T>& rhs )
{
	return *lhs == *rhs;
}

template<typename T>
inline bool operator==
	( const smart_constant<T>& lhs, const T& rhs )
{
	return *lhs == rhs;
}

//
// Function: operator<
//		r
//
//	smart_constant<T>  T ̔r`Ă܂B<,>,<=,>=
//	Zq́Aboost::operators ɂ莩܂B
//

template<typename T>
inline bool operator<
	( const smart_constant<T>& lhs, const smart_constant<T>& rhs )
{
	return *lhs < *rhs;
}

template<typename T>
inline bool operator<
	( const smart_constant<T>& lhs, const T& rhs )
{
	return *lhs < rhs;
}

template<typename T>
inline bool operator>
	( const smart_constant<T>& lhs, const T& rhs )
{
	return *lhs > rhs;
}



//=========================================================================
} // namespace end
