#include "stdafx.h"

//=========================================================================
// File: wininet++.cpp
//		wininet++̎łB
//=========================================================================

#include "wininet++.h"
using namespace wipp;



//-------------------------------------------------------------------------

Time::Time()
{
	::GetSystemTime(this);
}

string Time::toReadableString() const
{
	return (format("%04d-%02d-%02d  %02d:%02d")
		% wYear % wMonth % wDay % wHour % wMinute).str();
}

string Time::toRFC1123String() const
{
	// rfc1123-date ::= wkday "," SP date SP time SP "GMT"
	// date         ::= dd SP month SP yyyy
	//   dd,yyyy    ::= 2,4DIGIT
	// time         ::= hh ":" mm ":" ss
	//   hh,mm,ss   ::= 2DIGIT
	// wkday        ::= "Mon" | "Tue" | "Wed" | "Thu" | "Fri" | "Sat" | "Sun"
	// month        ::= "Jan" | "Feb" | "Mar" | "Apr" | "May" | "Jun"
	//                | "Jul" | "Aug" | "Sep" | "Oct" | "Nov" | "Dec"

	static const char* wkday[7] =
		{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	static const char* month[12] =
		{"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		 "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

	return (format("%s, %02d %s %04d %02d:%02d:%02d GMT")
		% wkday[wDayOfWeek%7] % wDay % month[(wMonth-1)%12] % wYear
		% wHour % wMinute % wSecond).str();
}

Time Time::UTCtoLOCAL() const
{
	Time t;
	FILETIME ft_u, ft_l;
	::SystemTimeToFileTime(this, &ft_u);
	::FileTimeToLocalFileTime(&ft_u, &ft_l);
	::FileTimeToSystemTime(&ft_l, &t);
	return t;
}

Time Time::LOCALtoUTC() const
{
	Time t;
	FILETIME ft_u, ft_l;
	::SystemTimeToFileTime(this, &ft_l);
	::LocalFileTimeToFileTime(&ft_l, &ft_u);
	::FileTimeToSystemTime(&ft_u, &t);
	return t;
}

bool Time::operator<(const Time& rhs) const
{
	#define CMP(mem) if(mem < rhs.mem) return true; if(rhs.mem < mem) return false

		CMP(wYear);
		CMP(wMonth);
		CMP(wDay);
		CMP(wHour);
		CMP(wMinute);
		CMP(wSecond);
		CMP(wMilliseconds);
		return false;

	#undef CMP
}

bool Time::operator==(const Time& rhs) const
{
	return wYear         == rhs.wYear
	    && wMonth        == rhs.wMonth
	    && wDay          == rhs.wDay
	    && wHour         == rhs.wHour
	    && wMinute       == rhs.wMinute
	    && wSecond       == rhs.wSecond
	    && wMilliseconds == rhs.wMilliseconds;
}

namespace wipp {
ostream& operator<<( ostream& os, const Time& time )
{
	return os
		<< time.wYear      << ':'
		<< time.wMonth     << ':'
		<< time.wDayOfWeek << ':'
		<< time.wDay       << ':'
		<< time.wHour      << ':'
		<< time.wMinute    << ':'
		<< time.wSecond;
}

istream& operator>>( istream& is, Time& time )
{
	is >> time.wYear;      is.get();
	is >> time.wMonth;     is.get();
	is >> time.wDayOfWeek; is.get();
	is >> time.wDay;       is.get();
	is >> time.wHour;      is.get();
	is >> time.wMinute;    is.get();
	is >> time.wSecond;
	return is;
}
}

//-------------------------------------------------------------------------

Agent::Agent( const char* name )
	: han( ::InternetOpen(name, INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0) )
{
	if( !han )
		throw wipp_exception("Agent Construction Failed");
}

Session Agent::connectTo( const char* servername, INTERNET_PORT port, int service )
{
	return Session(*this, servername, port, service);
}

//-------------------------------------------------------------------------

Session::Session(
	const Agent&  agent,
	const char*   servername,
	INTERNET_PORT port,
	int           service
)
	: agent     ( agent )
	, han       ( ::InternetConnect(agent.getTemporaryHandle(),
	                                servername, port, NULL, NULL, service, 0, 0) )
	, servername( servername )
{
	if( !han )
		throw wipp_exception("Session Construction Failed");
	setTimeOut(20);
}

HttpResponse Session::httpHEAD(const char* rsrcname, const Session::Headers& hdr)
{
	return makeHttpReq("HEAD", rsrcname, hdr);
}

HttpResponse Session::httpGET(const char* rsrcname, const Session::Headers& hdr)
{
	return makeHttpReq("GET", rsrcname, hdr);
}

void Session::setTimeOut(double sec)
{
	DWORD dw = static_cast<DWORD>(sec*1000);
	::InternetSetOption( *han, INTERNET_OPTION_RECEIVE_TIMEOUT, &dw, sizeof(dw) );
}

HttpResponse Session::makeHttpReq(
	const char* typ, const char* rsrc, const Session::Headers& hdr )
{
	// v쐬
	handle_internet hreq( ::HttpOpenRequest( *han, typ, rsrc, "HTTP/1.1",
		("http://"+servername+rsrc).c_str(), NULL, 0, 0 ) );

	if( !han )
		throw wipp_exception("Request Construction Failed");

	// ǉwb_đM
	string hdrstr;
	for(Headers::const_iterator i=hdr.begin(), e=hdr.end(); i!=e; ++i)
	{
		hdrstr += i->first;
		hdrstr += ": ";
		hdrstr += i->second;
		hdrstr += "\r\n";
	}
	if( !::HttpSendRequest(*hreq, hdrstr.c_str(), DWORD(-1), NULL, 0) )
		throw wipp_exception("Request Sending Failed");

	return HttpResponse( *this, hreq );
}

//-------------------------------------------------------------------------

HttpResponse::HttpResponse( const Session& session, handle_internet han )
	: session( session )
	, han    ( han )
{
}

bool HttpResponse::query( DWORD flag, void* buf, DWORD* siz )
{
	return FALSE != ::HttpQueryInfo( *han, flag, buf, siz, NULL );
}

int HttpResponse::status_code()
{
	DWORD data;
	DWORD size = sizeof(data);
	if( !query( HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &data, &size) )
		throw wipp_exception("status_code failed");
	return static_cast<int>(data);
}

int HttpResponse::status_code_nothrow()
{
	try {
		return status_code();
	} catch(const wipp_exception&) {
		return 404;
	}
}

ulong HttpResponse::content_length()
{
	DWORD data;
	DWORD size = sizeof(data);
	if( !query( HTTP_QUERY_CONTENT_LENGTH | HTTP_QUERY_FLAG_NUMBER, &data, &size) )
		throw wipp_exception("content_length failed");
	return static_cast<ulong>(data);
}

optional<ulong> HttpResponse::content_length_nothrow()
{
	try {
		return optional<ulong>( content_length() );
	} catch(const wipp_exception&) {
		return optional<ulong>();
	}
}

Time HttpResponse::last_modified()
{
	SYSTEMTIME data[100];
	DWORD size = sizeof(data);
	if( !query(HTTP_QUERY_LAST_MODIFIED | HTTP_QUERY_FLAG_SYSTEMTIME, &data, &size) )
		throw wipp_exception("last_modified failed");
	return data[0];
}

optional<Time> HttpResponse::last_modified_nothrow()
{
	try {
		return optional<Time>( last_modified() );
	} catch(const wipp_exception&) {
		return optional<Time>();
	}
}

string HttpResponse::query( DWORD flag )
{
	vector<char> buf(100); // magic number 100

	bool read_ok = false;
	while( !read_ok )
	{
		DWORD size = static_cast<DWORD>(buf.size());
		read_ok = query(flag, &buf[0], &size);
		if( !read_ok )
			if( ERROR_INSUFFICIENT_BUFFER==GetLastError() )
				buf.resize(size);
			else
				throw wipp_exception("query failed");
		else
			buf.resize(size);
	}
	return string(buf.begin(), buf.end());
}

optional<string> HttpResponse::query_nothrow( DWORD flag )
{
	try {
		return optional<string>( query(flag) );
	} catch(const wipp_exception&) {
		return optional<string>();
	}
}

vector<char> HttpResponse::readWhole()
{
	vector<char> data(100); // magic number 100
	unsigned idx = 0;

	bool read_end = false;
	while( !read_end )
	{
		DWORD read = 0;
		if( !::InternetReadFile( *han, &data[idx], DWORD(data.size()-idx), &read ) )
			throw wipp_exception("read failed");
		idx += read;
		if( read > 0 )
		{
			if( idx == data.size() )
				data.resize( data.size()*2 );
		}
		else
		{
			data.resize(idx);
			read_end = true;
		}
	}

	return data;
}

vector<char> HttpResponse::readWhole_nothrow()
{
	try {
		return readWhole();
	} catch(const wipp_exception&) {
		return vector<char>();
	}
}
